/*
 * This file is a part of the CaosDB Project.
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 Florian Spreckelsen <f.spreckelsen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program. If not, see
 * <https://www.gnu.org/licenses/>.
 *
 */
#include "ccaosdb.h"
#include "caosdb/connection.h"
#include "caosdb/constants.h"
#include "caosdb/utility.h"
#include "caosdb/status_code.h"
#include "caosdb/logging.h"
#include <cassert>
#include <exception>
#include <iostream>
#include <stdio.h>
#include <string.h>
#include <vector>

extern "C" {

#define CCAOSDB_LOGGER_NAME "ccaosdb"

/*
 * Macro for wrapping every function into a try-catch clause. If an exception
 * occurs, the given StatusCode is being returned.
 */
#define ERROR_RETURN_CODE(code, fun, body)                                     \
  fun {                                                                        \
    try {                                                                      \
      body                                                                     \
    } catch (const std::exception &exc) {                                      \
      caosdb::logging::caosdb_log_fatal(CCAOSDB_LOGGER_NAME, exc.what());      \
      return caosdb::StatusCode::code;                                         \
    }                                                                          \
  }

int caosdb_constants_LIBCAOSDB_VERSION_MAJOR() {
  return caosdb::LIBCAOSDB_VERSION_MAJOR;
}

int caosdb_constants_LIBCAOSDB_VERSION_MINOR() {
  return caosdb::LIBCAOSDB_VERSION_MINOR;
}

int caosdb_constants_LIBCAOSDB_VERSION_PATCH() {
  return caosdb::LIBCAOSDB_VERSION_PATCH;
}

int caosdb_constants_COMPATIBLE_SERVER_VERSION_MAJOR() {
  return caosdb::COMPATIBLE_SERVER_VERSION_MAJOR;
}

int caosdb_constants_COMPATIBLE_SERVER_VERSION_MINOR() {
  return caosdb::COMPATIBLE_SERVER_VERSION_MINOR;
}

int caosdb_constants_COMPATIBLE_SERVER_VERSION_PATCH() {
  return caosdb::COMPATIBLE_SERVER_VERSION_PATCH;
}

const char *caosdb_constants_COMPATIBLE_SERVER_VERSION_PRE_RELEASE() {
  return caosdb::COMPATIBLE_SERVER_VERSION_PRE_RELEASE;
}

const char *caosdb_utility_get_env_var(const char *name, const char *fallback) {
  return caosdb::utility::get_env_var(name, fallback);
}

const char *caosdb_get_status_description(int code) {
  return caosdb::get_status_description(code).c_str();
}

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_create_pem_file_certificate_provider(
                    caosdb_connection_certificate_provider *out,
                    const char *path),
                  {
                    out->wrapped_certificate_provider =
                      new caosdb::configuration::PemFileCertificateProvider(
                        std::string(path));
                    return 0;
                  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_connection_delete_certificate_provider(
    caosdb_connection_certificate_provider *provider),
  {
    delete static_cast<caosdb::configuration::CertificateProvider *>(
      provider->wrapped_certificate_provider);
    return 0;
  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_authentication_create_plain_password_authenticator(
                    caosdb_authentication_authenticator *out,
                    const char *username, const char *password),
                  {
                    out->wrapped_authenticator =
                      new caosdb::authentication::PlainPasswordAuthenticator(
                        std::string(username), std::string(password));
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_authentication_delete_authenticator(
                    caosdb_authentication_authenticator *authenticator),
                  {
                    delete static_cast<caosdb::authentication::Authenticator *>(
                      authenticator->wrapped_authenticator);
                    return 0;
                  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_connection_create_tls_connection_configuration(
    caosdb_connection_connection_configuration *out, const char *host,
    const int port, caosdb_authentication_authenticator *authenticator,
    caosdb_connection_certificate_provider *provider),
  {
    auto host_str = std::string(host);
    if (authenticator != nullptr && provider != nullptr) {
      auto wrapped_provider =
        static_cast<caosdb::configuration::CertificateProvider *>(
          provider->wrapped_certificate_provider);
      auto wrapped_authenticator =
        static_cast<caosdb::authentication::Authenticator *>(
          authenticator->wrapped_authenticator);
      out->wrapped_connection_configuration =
        new caosdb::configuration::TlsConnectionConfiguration(
          host_str, port, *wrapped_provider, *wrapped_authenticator);
    } else if (authenticator != nullptr) {
      auto wrapped_authenticator =
        static_cast<caosdb::authentication::Authenticator *>(
          authenticator->wrapped_authenticator);
      out->wrapped_connection_configuration =
        new caosdb::configuration::TlsConnectionConfiguration(
          host_str, port, *wrapped_authenticator);
    } else if (provider != nullptr) {
      auto wrapped_provider =
        static_cast<caosdb::configuration::CertificateProvider *>(
          provider->wrapped_certificate_provider);
      out->wrapped_connection_configuration =
        new caosdb::configuration::TlsConnectionConfiguration(
          host_str, port, *wrapped_provider);
    } else {
      out->wrapped_connection_configuration =
        new caosdb::configuration::TlsConnectionConfiguration(host_str, port);
    }
    return 0;
  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_connection_create_insecure_connection_configuration(
    caosdb_connection_connection_configuration *out, const char *host,
    const int port),
  {
    out->wrapped_connection_configuration =
      new caosdb::configuration::InsecureConnectionConfiguration(host, port);
    return 0;
  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_connection_delete_connection_configuration(
    caosdb_connection_connection_configuration *configuration),
  {
    delete static_cast<caosdb::configuration::ConnectionConfiguration *>(
      configuration->wrapped_connection_configuration);
    return 0;
  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_connection_create_connection(
    caosdb_connection_connection *out,
    const caosdb_connection_connection_configuration *configuration),
  {
    caosdb::configuration::ConnectionConfiguration *config =
      static_cast<caosdb::configuration::ConnectionConfiguration *>(
        configuration->wrapped_connection_configuration);
    out->wrapped_connection = new caosdb::connection::Connection(*config);
    return 0;
  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_delete_connection(
                    caosdb_connection_connection *connection),
                  {
                    delete static_cast<caosdb::connection::Connection *>(
                      connection->wrapped_connection);
                    return 0;
                  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_connection_get_version_info(
    caosdb_info_version_info *out,
    const caosdb_connection_connection *connection),
  {
    auto *wrapped_connection = static_cast<caosdb::connection::Connection *>(
      connection->wrapped_connection);

    auto status = wrapped_connection->RetrieveVersionInfoNoExceptions();
    if (status.IsError()) {
      return status.GetCode();
    }
    auto version_info = wrapped_connection->GetVersionInfo();

    out->major = (int)version_info->GetMajor();
    out->minor = (int)version_info->GetMinor();
    out->patch = (int)version_info->GetPatch();

    // copy pre_release, needs local variable because out->pre_release is const
    char *pre_release = (char *)malloc(
      sizeof(char) * (version_info->GetPreRelease().length() + 1));
    strcpy(pre_release, version_info->GetPreRelease().c_str());
    out->pre_release = pre_release;

    // copy build, needs local variable because out->build is const
    char *build =
      (char *)malloc(sizeof(char) * (version_info->GetBuild().length() + 1));
    strcpy(build, version_info->GetBuild().c_str());
    out->build = build;

    return 0;
  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_connection_connection_manager_get_default_connection(
    caosdb_connection_connection *out),
  {
    out->wrapped_connection =
      caosdb::connection::ConnectionManager::GetDefaultConnection().get();
    return 0;
  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_connection_manager_get_connection(
                    caosdb_connection_connection *out, const char *name),
                  {
                    out->wrapped_connection =
                      caosdb::connection::ConnectionManager::GetConnection(
                        std::string(name))
                        .get();
                    return 0;
                  })

/****************************************************************************
 * ENTITY STUFF AND TRANSACTIONS
 ****************************************************************************/
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_connection_create_transaction(
                    caosdb_connection_connection *connection,
                    caosdb_transaction_transaction *out),
                  {
                    auto *wrapped_connection =
                      static_cast<caosdb::connection::Connection *>(
                        connection->wrapped_connection);
                    out->wrapped_transaction =
                      wrapped_connection->CreateTransaction().release();
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_delete_transaction(
                    caosdb_transaction_transaction *transaction),
                  {
                    delete static_cast<caosdb::transaction::Transaction *>(
                      transaction->wrapped_transaction);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_transaction_retrieve_by_id(
                    caosdb_transaction_transaction *transaction,
                    const char *id),
                  {
                    auto *wrapped_transaction =
                      static_cast<caosdb::transaction::Transaction *>(
                        transaction->wrapped_transaction);
                    return wrapped_transaction->RetrieveById(std::string(id));
                  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_transaction_transaction_retrieve_by_ids(
    caosdb_transaction_transaction *transaction, const char *ids[]),
  {
    auto *wrapped_transaction = static_cast<caosdb::transaction::Transaction *>(
      transaction->wrapped_transaction);
    // Fill a string vector with the contents of the array of char arrays.
    std::vector<std::string> str_ids;
    for (const char **i = ids; *i; ++i) {
      str_ids.push_back(std::string(*i));
    }
    return wrapped_transaction->RetrieveById(str_ids.begin(), str_ids.end());
  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_transaction_query(
                    caosdb_transaction_transaction *transaction,
                    const char *query),
                  {
                    auto *wrapped_transaction =
                      static_cast<caosdb::transaction::Transaction *>(
                        transaction->wrapped_transaction);
                    return wrapped_transaction->Query(std::string(query));
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_transaction_execute(
                    caosdb_transaction_transaction *transaction),
                  {
                    auto *wrapped_transaction =
                      static_cast<caosdb::transaction::Transaction *>(
                        transaction->wrapped_transaction);
                    wrapped_transaction->Execute();
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_transaction_get_result_set(
                    caosdb_transaction_transaction *transaction,
                    caosdb_transaction_result_set *out),
                  {
                    auto *wrapped_transaction =
                      static_cast<caosdb::transaction::Transaction *>(
                        transaction->wrapped_transaction);
                    out->wrapped_result_set =
                      (void *)(&(wrapped_transaction->GetResultSet()));
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_transaction_get_count_result(
                    caosdb_transaction_transaction *transaction, long *out),
                  {
                    auto *wrapped_transaction =
                      static_cast<caosdb::transaction::Transaction *>(
                        transaction->wrapped_transaction);
                    long cr(wrapped_transaction->GetCountResult());
                    *out = cr;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_result_set_at(
                    caosdb_transaction_result_set *result_set,
                    caosdb_entity_entity *entity, int index),
                  {
                    // TODO(fspreck) how do we treat this? ResultSet
                    // is an abstract class, can we always fall back
                    // to MultiResultSet?
                    auto *wrapped_result_set =
                      static_cast<caosdb::transaction::MultiResultSet *>(
                        result_set->wrapped_result_set);
                    auto requested_entity = wrapped_result_set->At(index);
                    entity->wrapped_entity = (void *)(&requested_entity);
                    // TODO(fspreck) explicit pointers to id, name, ...
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_result_set_size(
                    caosdb_transaction_result_set *result_set, int *out),
                  {
                    auto *wrapped_result_set =
                      static_cast<caosdb::transaction::MultiResultSet *>(
                        result_set->wrapped_result_set);
                    int size(wrapped_result_set->Size());
                    *out = size;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_create_entity(caosdb_entity_entity *out), {
                    auto entity = caosdb::entity::Entity();
                    out->wrapped_entity = (void *)(&entity);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_delete_entity(caosdb_entity_entity *out), {
                    delete static_cast<caosdb::entity::Entity *>(
                      out->wrapped_entity);
                    return 0;
                  })
}
